import datetime
import _thread
import time
import traceback
from threading import Lock

import RPi.GPIO as GPIO
from iotgate.communicators.buses.spi import SPI
from iotgate.core.model.patterns import Singleton

__author__ = "diego"
__copyright__ = "Copyright 2016, Argotec SL"
__license__ = "?"
__version__ = "0.1.4-tfg"
__maintainer__ = "Diego Fidalgo"
__email__ = "dfidalgo@argotec.es"
__status__ = "Development"

VERBOSE = False
CHECK_INTERRUPT_INTERVAL = 0.1

# MRF24J40MC pins -> RPi GPIO Pins
RESET = 21
WAKE = 16
INTERRUPT = 24
STATE = 26
CHIP_SELECT = 7
# Interrupt signals values
RX_INTERRUPT_SIGNAL = 8

INTERVAL = 30


class MRF24J40MRadio(object, metaclass=Singleton):
    def __init__(self):
        self.spi = MRF24J40SPI()
        self.configuration = {}
        self.stop_flag = False
        self.interrupt = Lock()
        self.frame_register = {}
        self.server_request_func = None

    def init_rpi_gpio(self):
        GPIO.setmode(GPIO.BCM)
        # nReset
        GPIO.setup(RESET, GPIO.OUT)
        # GPIO0 WAKE from module
        GPIO.setup(WAKE, GPIO.OUT)
        # GPIO1 INTERRUPT from module
        GPIO.setup(INTERRUPT, GPIO.IN, pull_up_down=GPIO.PUD_DOWN)
        GPIO.add_event_detect(INTERRUPT, GPIO.FALLING, callback=self.treat_irq, bouncetime=10)
        # State
        GPIO.setup(STATE, GPIO.OUT)
        # Chip select
        GPIO.setup(CHIP_SELECT, GPIO.OUT)

    def get_all_values(self):
        return self.frame_register

    def reinit_radio(self):
        self.configure(self.configuration)

    def configure(self, configuration):
        self.configuration = configuration

    def configure_module(self):
        # Wake up!
        GPIO.output(WAKE, True)
        GPIO.output(CHIP_SELECT, False)
        # We reset the module
        GPIO.output(RESET, False)
        time.sleep(0.1)
        GPIO.output(RESET, True)
        # 0x07 - Perform a software Reset. Bit is auto-reset
        self.spi.write("SOFTRST", 0x07)
        # 0x98 - Initialize FIFOEN = 1 and TXONTS = 0x6.
        self.spi.write("PACON2", 0x98)
        # 0x95 - Initialize RFSTBL = 0x9.
        self.spi.write("TXSTBL", 0x95)
        # clean rx fifo
        self.spi.write("RXFLUSH", 0x01)
        # 11. BBREG2 (0x3A) = 0x80 - Set CCA mode to ED.
        self.spi.write("BBREG2", 0x80)
        # 12. CCAEDTH = 0x60 - Set CCA ED threshold.
        self.spi.write("CCAEDTH", 0x60)

        # 0x01 - Initialize VCOOPT = 0x02.
        self.spi.write("RFCON1", 0x01)
        # 0x80 - Enable PLL (PLLEN = 1).
        self.spi.write("RFCON2", 0x80)
        # 0x90 - Initialize TXFIL = 1 and 20MRECVR = 1.
        self.spi.write("RFCON6", 0x90)
        # 0x80 - Initialize SLPCLKSEL = 0x2 (100 kHz Internal oscillator).
        self.spi.write("RFCON7", 0x80)
        # 0x10 - Initialize RFVCO = 1.
        self.spi.write("RFCON8", 0x10)
        # 0x21 - Initialize CLKOUTEN = 1 and SLPCLKDIV = 0x01.
        self.spi.write("SLPCON0", 0x01)
        self.spi.write("SLPCON1", 0x20)
        # 13. BBREG6 (0x3E) = 0x40 - Set appended RSSI value to RXFIFO.
        self.spi.write("BBREG6", 0x40)
        ##
        # Interrupt config
        ##
        self.spi.write("INTCON", 0x00)
        # Set channel
        self.spi.write("RFCON0", channel[int(self.configuration['channel'])] | 0x02)
        self.spi.write("RFCTL", 0x04)
        self.spi.write("RFCTL", 0x00)

        time.sleep(0.1)
        # Transmission set to maximum level (0 dBm) (MC Version)
        self.spi.write("TRISGPIO", 0x08)
        self.spi.write("GPIO", 0x08)
        self.spi.write("RFCON3", 0x28)
        self.spi.write("TESTMODE", 0xf)
        # Set controller ID
        controller_id = str(self.configuration["controller_id"])
        high, low = int("0x" + controller_id[:2], 16), int("0x" + controller_id[2:], 16)
        self.spi.write("SADRL", low)
        self.spi.write("SADRH", high)
        # Set PANID
        pan_id = str(self.configuration["pan_id"])
        high, low = int("0x" + pan_id[:2], 16), int("0x" + pan_id[2:], 16)
        self.spi.write("PANIDL", low)
        self.spi.write("PANIDH", high)
        # Crear EADR0-7 extended address register
        for i in range(8):
            self.spi.write("EADR" + str(i), i)
        self.spi.write("RXMCR", 0x23)

        # 0x04 - Reset RF state machine.
        self.spi.write("RFCTL", 0x04)
        self.spi.write("RFCTL", 0x00)

    def set_request_func(self, func):
        self.server_request_func = func

    def treat_irq(self, gpio_number):
        int_reg = self.spi.read("INTSTAT")
        if int_reg == RX_INTERRUPT_SIGNAL:
            frame = self.get_rxfifo()
            device_id = self.read_frame(frame["rssi"], frame["bytes"])
            self.server_request_func(self.configuration, device_id)

    def get_rxfifo(self):
        with self.interrupt:
            self.spi.write("BBREG1", 0x04)
            mplusnplus2 = self.spi.read(0x300)
            read_bytes = []
            for i in range(mplusnplus2 + 2):
                read_bytes.append(self.spi.read(0x301 + i))
            self.spi.write("RXDECINV", 0)
            rssi = int(((read_bytes[-1] / 255.) * 70) - 90)
            data = dict(bytes=read_bytes[:-4], fcs=read_bytes[-4:-2], lqi=read_bytes[-2], rssi=rssi)
            # Allow to receive multiple message at same time
            self.spi.write("RXFLUSH", 0x61)
            self.spi.write("BBREG1", 0x00)
            time.sleep(0.150)
            # Clean FIFO
            self.spi.write("RXFLUSH", 0x01)

            # Reboot module
            self.reinit_radio()
        return data

    def read_message(self, device_id):
        if device_id in self.frame_register:
            if len(self.frame_register[device_id]) != 0:
                return self.frame_register[device_id].pop()
            else:
                return False
        else:
            return False

    def send_message(self, message, destiny):
        legacy = self.configuration.get('legacy', False)
        controller_id = (0x20, 0x00)
        destiny = (0x00, destiny)
        frame_bytes = [65, 200, 166, 69, 35, destiny[1], destiny[0], controller_id[1], controller_id[0],
                       0, 0, 0, 0, 0, 0]
        if legacy:
            frame_bytes.append(79)
        message = [ord(character) for character in message]
        frame_bytes += message
        if legacy:
            frame_bytes.append(79)

        self.write_txfifo(frame_bytes)

    def read_frame(self, rssi, frame):
        global INTERVAL
        try:
            ascii_frame = [str(chr(char)) for char in frame[16:]]
            device_id = frame[7]
        except Exception as err:
            if VERBOSE:
                print("[LLC-802.15.4 Server | " + str(datetime.datetime.now()) + "] Bad frame received. Exception: ",
                      err, frame)
            return ""
        message = ""
        for char in ascii_frame:
            message += char

        if device_id not in self.frame_register:
            self.frame_register[device_id] = []
        self.frame_register[device_id].append({"rssi": rssi, "message": message})

        return device_id

    def write_txfifo(self, bytes):
        while self.interrupt.locked():
            time.sleep(0.1)
        self.spi.write(0x000, len(bytes))
        self.spi.write(0x001, len(bytes))
        for i in range(2, len(bytes) + 2):
            self.spi.write(0x000 + i, bytes[i - 2])
        self.spi.write("TXNCON", 1)
        time.sleep(0.1)
        self.spi.write("RXFLUSH", 0x01)
        self.reinit_radio()

    def run(self):
        try:
            while not self.stop_flag:
                time.sleep(CHECK_INTERRUPT_INTERVAL)
        except Exception as err:
            if VERBOSE:
                print(traceback.format_exc())

    def start(self):
        self.init_rpi_gpio()
        self.spi.open()
        self.configure_module()
        _thread.start_new_thread(self.run, ())

    def stop(self):
        self.stop_flag = True
        self.spi.close()
        GPIO.cleanup()


class MRF24J40SPI(SPI):
    def __init__(self):
        # TODO especificar el bus y el device desde la configuración
        SPI.__init__(self, 0, 0)
        self.short_write_mask = 0x01
        self.long_write_mask = 0x10
        self.long_addr_bit = 0x80

    def _write_register(self, reg, value):
        """Writes the entire register, identified by its name  (datasheet)."""
        GPIO.output(CHIP_SELECT, False)
        if reg in short_addr_registers:
            reg_addr = short_addr_registers[reg]
            bytes = [((reg_addr << 1) + self.short_write_mask), value]
        else:
            if reg in long_addr_registers:
                reg_addr = long_addr_registers[reg]
            else:
                reg_addr = reg
            bytes = []
            bytes.append(((reg_addr >> 3) & 0x7F) + self.long_addr_bit)
            bytes.append(((reg_addr << 5) & 0xE0) + self.long_write_mask)
            bytes.append(value)

        self.transfer(bytes)
        GPIO.output(CHIP_SELECT, True)

    def _read_register(self, reg):
        """Reads the entire register, identified by its name (datasheet)."""
        GPIO.output(CHIP_SELECT, False)
        if reg in short_addr_registers:
            reg_addr = short_addr_registers[reg]
            bytes = [(reg_addr << 1), 0]
            result = self.transfer(bytes)
            GPIO.output(CHIP_SELECT, True)
            return result[1]
        else:
            if reg in long_addr_registers:
                reg_addr = long_addr_registers[reg]
            else:
                reg_addr = reg
            bytes = []
            bytes.append(((reg_addr >> 3) & 0x7F) + self.long_addr_bit)
            bytes.append((reg_addr << 5) & 0xE0)
            bytes.append(0)
            result = self.transfer(bytes)
            GPIO.output(CHIP_SELECT, True)
            return result[2]

    def _write_register_field(self, name, value):
        """Writes sub-register field, identified by its name (datasheet).
        These fields are defined in 'mrf24j40_registers' and can comprise of single bits, or
        several bits across multiple registers. For their names consult the datasheet."""
        reg_field_descr = register_fields[name]
        for dic in reg_field_descr:
            register = dic['register']
            shift = dic['shift']
            bits = dic['bits']
            reg_data = self._read_register(register)
            mask = sum([1 << i for i in range(0, bits)]) << shift
            reg_data &= (255 - mask)  # zero the masked range
            reg_data |= (value << shift) & mask
            self._write_register(register, reg_data)
            value = (value >> bits)  # shift the bits for this field away

    def _read_register_field(self, name):
        """Reads sub-register field, identified by its name (datasheet).
        These fields are defined in 'mrf24j40_registers' and can comprise of single bits, or
        several bits across multiple registers. For their names consult the datasheet."""
        reg_field_descr = register_fields[name]
        return_value = 0
        return_shift = 0
        for dic in reg_field_descr:
            register = dic['register']
            shift = dic['shift']
            bits = dic['bits']
            reg_data = self._read_register(register)
            mask = sum([1 << i for i in range(0, bits)]) << shift
            reg_data = reg_data & mask
            reg_data = reg_data >> shift
            return_value += reg_data << return_shift
            return_shift += bits
        return return_value

    def write(self, name, value):
        """Convenient function to write either to registers of sub-register fields."""
        if name in list(register_fields.keys()):
            self._write_register_field(name, value)
        else:
            self._write_register(name, value)

    def read(self, name):
        """Convenient function to read either to registers of sub-register fields."""
        if name in list(register_fields.keys()):
            return self._read_register_field(name)
        else:
            return self._read_register(name)


channel = {11: 0x00,
           12: 0x10,
           13: 0x20,
           14: 0x30,
           15: 0x40,
           16: 0x50,
           17: 0x60,
           18: 0x70,
           19: 0x80,
           20: 0x90,
           21: 0xa0,
           22: 0xb0,
           23: 0xc0,
           24: 0xd0,
           25: 0xe0,
           26: 0xf0}

short_addr_registers = {
    "RXMCR": 0x00,
    "PANIDL": 0x01,
    "PANIDH": 0x02,
    "SADRL": 0x03,
    "SADRH": 0x04,
    "EADR0": 0x05,
    "EADR1": 0x06,
    "EADR2": 0x07,
    "EADR3": 0x08,
    "EADR4": 0x09,
    "EADR5": 0x0A,
    "EADR6": 0x0B,
    "EADR7": 0x0C,
    "RXFLUSH": 0x0D,
    "ORDER": 0x10,
    "TXMCR": 0x11,
    "ACKTMOUT": 0x12,
    "ESLOTG1": 0x13,
    "SYMTICKL": 0x14,
    "SYMTICKH": 0x15,
    "PACON0": 0x16,
    "PACON1": 0x17,
    "PACON2": 0x18,
    "TXBCON0": 0x1A,
    "TXNCON": 0x1B,
    "TXG1CON": 0x1C,
    "TXG2CON": 0x1D,
    "ESLOTG23": 0x1E,
    "ESLOTG45": 0x1F,
    "ESLOTG67": 0x20,
    "TXPEND": 0x21,
    "WAKECON": 0x22,
    "FRMOFFSET": 0x23,
    "TXSTAT": 0x24,
    "TXBCON1": 0x25,
    "GATECLK": 0x26,
    "TXTIME": 0x27,
    "HSYMTMRL": 0x28,
    "HSYMTMRH": 0x29,
    "SOFTRST": 0x2A,
    "SECCON0": 0x2C,
    "SECCON1": 0x2D,
    "TXSTBL": 0x2E,
    "RXSR": 0x30,
    "INTSTAT": 0x31,
    "INTCON": 0x32,
    "GPIO": 0x33,
    "TRISGPIO": 0x34,
    "SLPACK": 0x35,
    "RFCTL": 0x36,
    "SECCR2": 0x37,
    "BBREG0": 0x38,
    "BBREG1": 0x39,
    "BBREG2": 0x3A,
    "BBREG3": 0x3B,
    "BBREG4": 0x3C,
    "BBREG6": 0x3E,
    "CCAEDTH": 0x3F}

long_addr_registers = {
    "RFCON0": 0x200,
    "RFCON1": 0x201,
    "RFCON2": 0x202,
    "RFCON3": 0x203,
    "RFCON5": 0x205,
    "RFCON6": 0x206,
    "RFCON7": 0x207,
    "RFCON8": 0x208,
    "SLPCAL0": 0x209,
    "SLPCAL1": 0x20A,
    "SLPCAL2": 0x20B,
    "RFSTATE": 0x20F,
    "RSSI": 0x210,
    "SLPCON0": 0x211,
    "SLPCON1": 0x220,
    "WAKETIMEL": 0x222,
    "WAKETIMEH": 0x223,
    "REMCNTL": 0x224,
    "REMCNTH": 0x225,
    "MAINCNT0": 0x226,
    "MAINCNT1": 0x227,
    "MAINCNT2": 0x228,
    "MAINCNT3": 0x229,
    "TESTMODE": 0x22F,
    "ASSOEADR0": 0x230,
    "ASSOEADR1": 0x231,
    "ASSOEADR2": 0x232,
    "ASSOEADR3": 0x233,
    "ASSOEADR4": 0x234,
    "ASSOEADR5": 0x235,
    "ASSOEADR6": 0x236,
    "ASSOEADR7": 0x237,
    "ASSOSADR0": 0x238,
    "ASSOSADR1": 0x239,
    "UPNONCE0": 0x240,
    "UPNONCE1": 0x241,
    "UPNONCE2": 0x242,
    "UPNONCE3": 0x243,
    "UPNONCE4": 0x244,
    "UPNONCE5": 0x245,
    "UPNONCE6": 0x246,
    "UPNONCE7": 0x247,
    "UPNONCE8": 0x248,
    "UPNONCE9": 0x249,
    "UPNONCE10": 0x24A,
    "UPNONCE11": 0x24B,
    "UPNONCE12": 0x24C}

register_fields = {
    '20MRECVR': [{'shift': 4, 'register': 'RFCON6', 'bits': 1}],
    'BATEN': [{'shift': 3, 'register': 'RFCON6', 'bits': 1}],
    'BATIND': [{'shift': 5, 'register': 'RXSR', 'bits': 1}],
    'BATLIFEXT': [{'shift': 6, 'register': 'TXMCR', 'bits': 1}],
    'BATTH': [{'shift': 4, 'register': 'RFCON5', 'bits': 4}],
    'BCNONLY': [{'shift': 1, 'register': 'RXFLUSH', 'bits': 1}],
    'BO': [{'shift': 4, 'register': 'ORDER', 'bits': 4}],
    'CAP': [{'shift': 0, 'register': 'ESLOTG1', 'bits': 4}],
    'CCACSTH': [{'shift': 2, 'register': 'BBREG2', 'bits': 4}],
    'CCAEDTH': [{'shift': 0, 'register': 'CCAEDTH', 'bits': 8}],
    'CCAFAIL': [{'shift': 5, 'register': 'TXSTAT', 'bits': 1}],
    'CCAMODE': [{'shift': 6, 'register': 'BBREG2', 'bits': 2}],
    'CHANNEL': [{'shift': 4, 'register': 'RFCON0', 'bits': 4}],
    'CLKOUTEN': [{'shift': 5, 'register': 'SLPCON1', 'bits': 1}],
    'CLKOUTMODE': [{'shift': 0, 'register': 'RFCON7', 'bits': 2}],
    'CMDONLY': [{'shift': 3, 'register': 'RXFLUSH', 'bits': 1}],
    'COORD': [{'shift': 2, 'register': 'RXMCR', 'bits': 1}],
    'CSMABF': [{'shift': 0, 'register': 'TXMCR', 'bits': 3}],
    'CSTH': [{'shift': 5, 'register': 'BBREG4', 'bits': 3}],
    'DATAONLY': [{'shift': 2, 'register': 'RXFLUSH', 'bits': 1}],
    'DISDEC': [{'shift': 1, 'register': 'SECCON1', 'bits': 1}],
    'DISENC': [{'shift': 0, 'register': 'SECCON1', 'bits': 1}],
    'DRPACK': [{'shift': 7, 'register': 'ACKTMOUT', 'bits': 1}],
    'ERRPKT': [{'shift': 1, 'register': 'RXMCR', 'bits': 1}],
    'FIFOEN': [{'shift': 7, 'register': 'PACON2', 'bits': 1}],
    'FPACK': [{'shift': 0, 'register': 'TXPEND', 'bits': 1}],
    'FPSTAT': [{'shift': 4, 'register': 'TXNCON', 'bits': 1}],
    'GPIO': [{'shift': 0, 'register': 'GPIO', 'bits': 6}],
    'GTS1': [{'shift': 4, 'register': 'ESLOTG1', 'bits': 4}],
    'GTS2': [{'shift': 0, 'register': 'ESLOTG23', 'bits': 4}],
    'GTS3': [{'shift': 4, 'register': 'ESLOTG23', 'bits': 4}],
    'GTS4': [{'shift': 0, 'register': 'ESLOTG45', 'bits': 4}],
    'GTS5': [{'shift': 4, 'register': 'ESLOTG45', 'bits': 4}],
    'GTS6': [{'shift': 0, 'register': 'ESLOTG67', 'bits': 4}],
    'GTSON': [{'shift': 3, 'register': 'GATECLK', 'bits': 1}],
    'GTSSWITCH': [{'shift': 1, 'register': 'TXPEND', 'bits': 1}],
    'HSYMTMR': [{'shift': 0, 'register': 'HSYMTMRL', 'bits': 8}, {'shift': 2, 'register': 'HSYMTMRH', 'bits': 6}],
    'HSYMTMRIE': [{'shift': 5, 'register': 'INTCON', 'bits': 1}],
    'HSYMTMRIF': [{'shift': 5, 'register': 'INTSTAT', 'bits': 1}],
    'IMMWAKE': [{'shift': 7, 'register': 'WAKECON', 'bits': 1}],
    'INDIRECT': [{'shift': 3, 'register': 'TXNCON', 'bits': 1}],
    'INTEDGE': [{'shift': 1, 'register': 'SLPCON0', 'bits': 1}],
    'INTL': [{'shift': 0, 'register': 'WAKECON', 'bits': 1}],
    'MACMINBE': [{'shift': 3, 'register': 'TXMCR', 'bits': 2}],
    'MAINCNT': [{'shift': 0, 'register': 'MAINCNT0', 'bits': 8}, {'shift': 0, 'register': 'MAINCNT1', 'bits': 8},
                {'shift': 0, 'register': 'MAINCNT2', 'bits': 8}, {'shift': 0, 'register': 'MAINCNT3', 'bits': 2}],
    'MAWD': [{'shift': 0, 'register': 'ACKTMOUT', 'bits': 7}],
    'MLIFS': [{'shift': 2, 'register': 'TXPEND', 'bits': 6}],
    'MSIFS': [{'shift': 0, 'register': 'TXSTBL', 'bits': 4}],
    'NOACKRSP': [{'shift': 5, 'register': 'RXMCR', 'bits': 1}],
    'NOCSMA': [{'shift': 7, 'register': 'TXMCR', 'bits': 1}],
    'OFFSET': [{'shift': 0, 'register': 'FRMOFFSET', 'bits': 8}],
    'PANCOORD': [{'shift': 3, 'register': 'RXMCR', 'bits': 1}],
    'PANID': [{'shift': 0, 'register': 'PANIDL', 'bits': 8}, {'shift': 0, 'register': 'PANIDH', 'bits': 8}],
    'PAONT': [{'shift': 0, 'register': 'PACON0', 'bits': 8}],
    'PAONTS': [{'shift': 1, 'register': 'PACON1', 'bits': 4}],
    'PLLEN': [{'shift': 7, 'register': 'RFCON2', 'bits': 1}],
    'PRECNT': [{'shift': 2, 'register': 'BBREG4', 'bits': 3}],
    'PREDETTH': [{'shift': 1, 'register': 'BBREG3', 'bits': 3}],
    'PREVALIDTH': [{'shift': 4, 'register': 'BBREG3', 'bits': 4}],
    'PROMI': [{'shift': 0, 'register': 'RXMCR', 'bits': 1}],
    'REGWAKE': [{'shift': 6, 'register': 'WAKECON', 'bits': 1}],
    'REMCNT': [{'shift': 0, 'register': 'REMCNTL', 'bits': 8}, {'shift': 0, 'register': 'REMCNTH', 'bits': 8}],
    'RFOPT': [{'shift': 0, 'register': 'RFCON0', 'bits': 4}],
    'RFSTATE': [{'shift': 5, 'register': 'RFSTATE', 'bits': 3}],
    'RFSTBL': [{'shift': 4, 'register': 'TXSTBL', 'bits': 4}],
    'RFVCO': [{'shift': 4, 'register': 'RFCON8', 'bits': 1}],
    'RSSI': [{'shift': 0, 'register': 'RSSI', 'bits': 8}],
    'RSSINUM': [{'shift': 4, 'register': 'TXBCON1', 'bits': 2}],
    'RSSIWAIT': [{'shift': 3, 'register': 'TESTMODE', 'bits': 2}],
    'RSTBB': [{'shift': 1, 'register': 'SOFTRST', 'bits': 1}],
    'RSTMAC': [{'shift': 0, 'register': 'SOFTRST', 'bits': 1}],
    'RSTPWR': [{'shift': 2, 'register': 'SOFTRST', 'bits': 1}],
    'RXCIPHER': [{'shift': 3, 'register': 'SECCON0', 'bits': 3}],
    'RXDECINV': [{'shift': 2, 'register': 'BBREG1', 'bits': 1}],
    'RXFLUSH': [{'shift': 0, 'register': 'RXFLUSH', 'bits': 1}],
    'RXIE': [{'shift': 3, 'register': 'INTCON', 'bits': 1}],
    'RXIF': [{'shift': 3, 'register': 'INTSTAT', 'bits': 1}],
    'SECDECERR': [{'shift': 2, 'register': 'RXSR', 'bits': 1}],
    'SADR': [{'shift': 0, 'register': 'SADRL', 'bits': 8}, {'shift': 0, 'register': 'SADRH', 'bits': 8}],
    'SECIE': [{'shift': 4, 'register': 'INTCON', 'bits': 1}],
    'SECIF': [{'shift': 4, 'register': 'INTSTAT', 'bits': 1}],
    'SECIGNORE': [{'shift': 7, 'register': 'SECCON0', 'bits': 1}],
    'SECSTART': [{'shift': 6, 'register': 'SECCON0', 'bits': 1}],
    'SLOTTED': [{'shift': 5, 'register': 'TXMCR', 'bits': 1}],
    'SLPACK': [{'shift': 7, 'register': 'SLPACK', 'bits': 1}],
    'SLPCAL': [{'shift': 0, 'register': 'SLPCAL0', 'bits': 8}],
    'SLPCAL1': [{'shift': 2, 'register': 'SLPCAL1', 'bits': 6}],
    'SLPCALEN': [{'shift': 4, 'register': 'SLPCAL2', 'bits': 1}],
    'SLPCALRDY': [{'shift': 7, 'register': 'SLPCAL2', 'bits': 1}],
    'SLPCLKDIV': [{'shift': 0, 'register': 'SLPCON1', 'bits': 5}],
    'SLPCLKEN': [{'shift': 0, 'register': 'SLPCON0', 'bits': 1}],
    'SLPCLKSEL': [{'shift': 6, 'register': 'RFCON7', 'bits': 2}],
    'SLPIE': [{'shift': 7, 'register': 'INTCON', 'bits': 1}],
    'SLPIF': [{'shift': 7, 'register': 'INTSTAT', 'bits': 1}],
    'SO': [{'shift': 0, 'register': 'ORDER', 'bits': 4}],
    'STARTCNT': [{'shift': 7, 'register': 'MAINCNT3', 'bits': 1}],
    'TESTMODE': [{'shift': 0, 'register': 'TESTMODE', 'bits': 3}],
    'TICKP': [{'shift': 0, 'register': 'SYMTICKL', 'bits': 8}],
    'TRISGP': [{'shift': 0, 'register': 'TRISGPIO', 'bits': 6}],
    'TURBO': [{'shift': 0, 'register': 'BBREG0', 'bits': 1}],
    'TURNTIME': [{'shift': 4, 'register': 'TXTIME', 'bits': 4}],
    'TXBCIPHER': [{'shift': 4, 'register': 'SECCON1', 'bits': 3}],
    'TXBMSK': [{'shift': 7, 'register': 'TXBCON1', 'bits': 1}],
    'TXBSECEN': [{'shift': 1, 'register': 'TXBCON0', 'bits': 1}],
    'TXBTRIG': [{'shift': 0, 'register': 'TXBCON0', 'bits': 1}],
    'TXFIL': [{'shift': 7, 'register': 'RFCON6', 'bits': 1}],
    'TXG1ACKREQ': [{'shift': 2, 'register': 'TXG1CON', 'bits': 1}],
    'TXG1CIPHER': [{'shift': 0, 'register': 'SECCR2', 'bits': 3}],
    'TXG1FNT': [{'shift': 3, 'register': 'TXSTAT', 'bits': 1}],
    'TXG1IE': [{'shift': 1, 'register': 'INTCON', 'bits': 1}],
    'TXG1IF': [{'shift': 1, 'register': 'INTSTAT', 'bits': 1}],
    'TXG1RETRY': [{'shift': 6, 'register': 'TXG1CON', 'bits': 2}],
    'TXG1SECEN': [{'shift': 1, 'register': 'TXG1CON', 'bits': 1}],
    'TXG1SLOT': [{'shift': 3, 'register': 'TXG1CON', 'bits': 3}],
    'TXG1STAT': [{'shift': 1, 'register': 'TXSTAT', 'bits': 1}],
    'TXG1TRIG': [{'shift': 0, 'register': 'TXG1CON', 'bits': 1}],
    'TXG2ACKREQ': [{'shift': 2, 'register': 'TXG2CON', 'bits': 1}],
    'TXG2CIPHER': [{'shift': 3, 'register': 'SECCR2', 'bits': 3}],
    'TXG2FNT': [{'shift': 4, 'register': 'TXSTAT', 'bits': 1}],
    'TXG2IE': [{'shift': 2, 'register': 'INTCON', 'bits': 1}],
    'TXG2IF': [{'shift': 2, 'register': 'INTSTAT', 'bits': 1}],
    'TXG2RETRY': [{'shift': 6, 'register': 'TXG2CON', 'bits': 2}],
    'TXG2SECEN': [{'shift': 1, 'register': 'TXG2CON', 'bits': 1}],
    'TXG2SLOT': [{'shift': 3, 'register': 'TXG2CON', 'bits': 3}],
    'TXG2STAT': [{'shift': 2, 'register': 'TXSTAT', 'bits': 1}],
    'TXG2TRIG': [{'shift': 0, 'register': 'TXG2CON', 'bits': 1}],
    'TXNACKREQ': [{'shift': 2, 'register': 'TXNCON', 'bits': 1}],
    'TXNCIPHER': [{'shift': 0, 'register': 'SECCON0', 'bits': 3}],
    'TXNIE': [{'shift': 0, 'register': 'INTCON', 'bits': 1}],
    'TXNIF': [{'shift': 0, 'register': 'INTSTAT', 'bits': 1}],
    'TXNRETRY': [{'shift': 6, 'register': 'TXSTAT', 'bits': 2}],
    'TXNSECEN': [{'shift': 1, 'register': 'TXNCON', 'bits': 1}],
    'TXNSTAT': [{'shift': 0, 'register': 'TXSTAT', 'bits': 1}],
    'TXNTRIG': [{'shift': 0, 'register': 'TXNCON', 'bits': 1}],
    'TXONT': [{'shift': 1, 'register': 'SYMTICKH', 'bits': 7}],
    'TXONTS': [{'shift': 2, 'register': 'PACON2', 'bits': 4}],
    'TXPWRL': [{'shift': 6, 'register': 'RFCON3', 'bits': 2}],
    'TXPWRS': [{'shift': 3, 'register': 'RFCON3', 'bits': 3}],
    'UPDEC': [{'shift': 7, 'register': 'SECCR2', 'bits': 1}],
    'UPENC': [{'shift': 6, 'register': 'SECCR2', 'bits': 1}],
    'UPSECERR': [{'shift': 6, 'register': 'RXSR', 'bits': 1}],
    'VCOOPT': [{'shift': 0, 'register': 'RFCON1', 'bits': 8}],
    'WAKECNT': [{'shift': 0, 'register': 'SLPACK', 'bits': 7}],
    'WAKEIE': [{'shift': 6, 'register': 'INTCON', 'bits': 1}],
    'WAKEIF': [{'shift': 6, 'register': 'INTSTAT', 'bits': 1}],
    'WAKEPAD': [{'shift': 5, 'register': 'RXFLUSH', 'bits': 1}],
    'WAKEPOL': [{'shift': 6, 'register': 'RXFLUSH', 'bits': 1}],
    'WAKETIME': [{'shift': 0, 'register': 'WAKETIMEL', 'bits': 8}, {'shift': 0, 'register': 'WAKETIMEH', 'bits': 3}],
    'WU/BCN': [{'shift': 6, 'register': 'TXBCON1', 'bits': 1}]
}
