#!/usr/bin/env python

from iotgate.core.configurations import LoggerConfiguration
from iotgate.core.configurations import PersistenceConfiguration
from iotgate.core.logger import Logger
from iotgate.core.model import DatabaseManager
from iotgate.core.model.managers import AdaptorManager
from iotgate.core.model.managers import CommunicatorManager
from iotgate.core.model.managers import NodeManager
from iotgate.core.model.managers import ProductCatalog
from iotgate.core.model.nodes import NodeUser
from iotgate.core.model.patterns import Subject
from iotgate.deployment.helpers import AdaptorManagerHelper
from iotgate.deployment.helpers import CommunicatorManagerHelper
from iotgate.deployment.helpers import NodeManagerHelper
from iotgate.deployment.helpers import ProductCatalogHelper

__author__ = "diego"
__copyright__ = "Copyright 2016, Argotec SL"
__license__ = "?"
__version__ = "0.1.4-tfg"
__maintainer__ = "Diego Fidalgo"
__email__ = "dfidalgo@argotec.es"
__status__ = "Development"


DBManager = None


class Engine(Subject):
    VERSION = 'v0.1.4'

    def __init__(self, configuration_path, reset_database=False, before_start=None, after_start=None):
        Subject.__init__(self)
        self.work = True
        self.configuration_path = configuration_path

        # We prepare the environment
        self._create_logger_service()
        self.logger.info('Middleware %s' % Engine.VERSION)

        self._create_persistence_management()
        if reset_database:
            self._reset_database()

        self._create_model_managers()
        self._create_deployment_helpers()

        if before_start:
            before_start(self)

        self.communicator_manager.start_communicators()

        if after_start:
            after_start(self)

    def _create_logger_service(self):
        lc = LoggerConfiguration(self.configuration_path).get_logger_configuration()
        self.logger = Logger(lc)

    def _create_persistence_management(self):
        global DBManager

        pc = PersistenceConfiguration(self.configuration_path).get_engine_configuration()
        self._database_manager = DatabaseManager(pc)

        DBManager = self.database_manager
        self.session = self.database_manager.session

        self.logger.info("Persistence management ready: " + str(pc))

    def _reset_database(self):
        self.database_manager.drop_schema()
        self.database_manager.create_schema()

    def _create_model_managers(self):
        self.logger.info("Loading adaptors and communicators...")

        self._communicator_manager = CommunicatorManager(self.database_manager)
        self.logger.info("Communicator manager loaded.")

        self._adaptor_manager = AdaptorManager(self.database_manager, self.communicator_manager)
        self.logger.info("Adaptor manager loaded.")

        self.logger.info("Loading products...")

        self._product_manager = ProductCatalog(self.database_manager)
        self.logger.info("Product catalog loaded.")
        self.logger.info("Loading node groups and nodes...")

        self._node_manager = NodeManager(self.database_manager, self.adaptor_manager)

        self.logger.info("Node groups and nodes created!")

    def _create_deployment_helpers(self):
        self.product_catalog_helper = ProductCatalogHelper(self.product_manager)
        self.communicator_manager_helper = CommunicatorManagerHelper(self.communicator_manager)
        self.adaptor_manager_helper = AdaptorManagerHelper(self.adaptor_manager)
        self.node_manager_helper = NodeManagerHelper(self.node_manager)

    @staticmethod
    def get_version() -> str:
        return Engine.VERSION

    @property
    def database_manager(self):
        return self._database_manager

    @property
    def communicator_manager(self) -> CommunicatorManager:
        return self._communicator_manager

    @property
    def adaptor_manager(self) -> AdaptorManager:
        return self._adaptor_manager

    @property
    def product_manager(self) -> ProductCatalog:
        return self._product_manager

    @property
    def node_manager(self) -> NodeManager:
        return self._node_manager

    def get_user(self, username: str) -> NodeUser:
        return self.session.query(NodeUser).filter_by(username=username).one()

    def get_workers(self) -> dict:
        return {
            "CommunicatorManager": self.communicator_manager_helper,
            "AdaptorManager": self.adaptor_manager_helper,
            "ProductCatalog": self.product_catalog_helper,
            "NodeManager": self.node_manager_helper
        }

    def safe_close(self):
        self.communicator_manager.stop_communicators()
        self.session.remove()
        self.work = False
