from abc import ABCMeta

from sqlalchemy.engine import create_engine
from sqlalchemy.ext.declarative.api import declarative_base
from sqlalchemy.ext.declarative.api import DeclarativeMeta
from sqlalchemy.sql.schema import Table
from sqlalchemy.sql.schema import Column
from sqlalchemy.sql.schema import ForeignKey
from sqlalchemy.sql.sqltypes import Integer
from sqlalchemy.orm.scoping import scoped_session
from sqlalchemy.orm.session import sessionmaker

from iotgate.core.model.patterns import Singleton

__author__ = "diego"
__copyright__ = "Copyright 2015, Argotec SL"
__license__ = "?"
__version__ = "0.1"
__maintainer__ = "Diego Fidalgo"
__email__ = "dfidalgo@argotec.es"
__status__ = "Development"


class _Base(object):
    """Base class for SQLAlchemy model classes."""
    __table_args__ = {'mysql_engine': 'InnoDB', 'mysql_charset': 'utf8'}


Base = declarative_base(cls=_Base)


class DatabaseManager(object, metaclass=Singleton):
    def __init__(self, configuration):
        self._configuration = configuration
        if configuration["engine"].split(":")[0] == "sqlite":
            self._engine = create_engine(configuration["engine"])
        else:
            self._engine = create_engine(configuration["engine"], pool_size=20, max_overflow=0)

        self._Session = scoped_session(sessionmaker(bind=self.engine, expire_on_commit=False))

    @property
    def configuration(self):
        return self._configuration

    @property
    def engine(self):
        return self._engine

    @property
    def session(self):
        return self._Session

    def create_schema(self):
        Base.metadata.create_all(self.engine, checkfirst=True)

    def drop_schema(self):
        Base.metadata.drop_all(self.engine)

    @staticmethod
    def save_states(states: list):
        from iotgate.core.main import DBManager
        from iotgate.core.model.nodes import NodeComponentState

        insert_query = NodeComponentState.__table__.insert().prefix_with('OR REPLACE')
        connection = DBManager.engine.connect()
        try:
            connection.execute(insert_query, states)
        finally:
            connection.close()


adaptor_communicator_table = Table('adaptor_communicator', Base.metadata,
                                   Column('adaptor_id', Integer, ForeignKey('adaptor.id')),
                                   Column('communicator_id', Integer, ForeignKey('communicator.id'))
                                   )

product_component_table = Table('product_product_component', Base.metadata,
                                Column('product_id', Integer, ForeignKey('product.id')),
                                Column('product_component_id', Integer, ForeignKey('product_component.id'))
                                )


class AbstractBase(DeclarativeMeta, ABCMeta):
    pass
