from abc import abstractmethod
from sqlalchemy.orm import relationship
from sqlalchemy.sql.schema import Column
from sqlalchemy.sql.sqltypes import String

from iotgate.core.model import Base
from iotgate.core.model import AbstractBase
from iotgate.core.model import adaptor_communicator_table
from iotgate.core.model.communicators import Communicator
from iotgate.core.model.nodes import Actuator
from iotgate.core.model.patterns import Subject
from iotgate.core.model.products import ProductActuatorAction

__author__ = "diego"
__copyright__ = "Copyright 2016, Argotec SL"
__license__ = "?"
__version__ = "0.1.4-tfg"
__maintainer__ = "Diego Fidalgo"
__email__ = "dfidalgo@argotec.es"
__status__ = "Development"


class Adaptor(Base, Subject, metaclass=AbstractBase):
    """
    Events
        info: called when you want to provide some info (debug purposes).
            param - message: str
        error: called for any error occurred.
            param - message: str => Error message.
    """    
    event_types = ['info', 'error']
    __tablename__ = 'adaptor'

    id = Column(String(128), primary_key=True)
    type = Column(String(128), nullable=False)
    communicators = relationship('Communicator', secondary=adaptor_communicator_table, enable_typechecks=False)
    nodes = relationship('Node', lazy='subquery')

    @staticmethod
    def create(adaptor_id: str, adaptor_type: str, communicators: list):
        module_name = ''
        for item in adaptor_type.split('.')[:-1]:
            module_name += '.' + item

        class_name = adaptor_type.split('.')[-1]
        module = __import__('iotgate.adaptors' + module_name, fromlist=[class_name])

        adaptor_class = getattr(module, class_name)
        adaptor = adaptor_class(id=adaptor_id, communicators=communicators)

        return adaptor

    def __init__(self, id: str, _class, communicators: list):
        Subject.__init__(self)
        self.id = id
        self.type = ''
        for word in _class.__module__.split(".")[-2:]:
            self.type += word + '.'
        self.type += _class.__name__
        self.communicators = communicators
        self.managed_nodes = {}
        self.configure()

    def get_id(self) -> str:
        return self.id

    def _set_id(self, id_: str):
        self.id = id_

    def get_type(self) -> str:
        return self.type

    def _set_type(self, type_: str):
        self.type = type_

    def get_communicator_by_type(self, type_: str) -> Communicator:
        return next((c for c in self.communicators if c.get_type() == type_))

    def add_managed_node(self, node):
        self.managed_nodes[node.get_network_id()] = node

    def get_managed_nodes_networks_ids(self):
        return self.managed_nodes.keys()

    def get_managed_node(self, network_id: str):
        return self.managed_nodes[network_id]

    def serialize(self):
        return {
            "id": self.id,
            "type": self.type,
            "communicators": [communicator.get_id() for communicator in self.communicators]
        }

    @abstractmethod
    def configure(self):
        pass

    @abstractmethod
    def send_action(self, actuator: Actuator, action: ProductActuatorAction, params):
        pass
