from abc import abstractmethod
from sqlalchemy.orm import relationship
from sqlalchemy.orm.collections import attribute_mapped_collection
from sqlalchemy.ext.associationproxy import association_proxy
from sqlalchemy.sql.schema import Column
from sqlalchemy.sql.schema import ForeignKey
from sqlalchemy.sql.sqltypes import String

from iotgate.core.model import Base
from iotgate.core.model import AbstractBase
from iotgate.core.model import adaptor_communicator_table
from iotgate.core.model.patterns import Subject

__author__ = "diego"
__copyright__ = "Copyright 2015, Argotec SL"
__license__ = "?"
__version__ = "0.1"
__maintainer__ = "Diego Fidalgo"
__email__ = "dfidalgo@argotec.es"
__status__ = "Development"


class Communicator(Base, Subject, metaclass=AbstractBase):
    __tablename__ = 'communicator'

    id = Column(String(128), primary_key=True)
    type = Column(String(128), nullable=False)
    _params = relationship('CommunicatorParam', collection_class=attribute_mapped_collection("key"))
    configuration = association_proxy("_params", "value", creator=lambda k, v: CommunicatorParam(key=k, value=v))
    adaptors = relationship('Adaptor', secondary=adaptor_communicator_table, enable_typechecks=False)


    @staticmethod
    def create(comm_id: str, comm_type: str, configuration: dict):
        module_name = ''
        for item in comm_type.split(".")[:-1]:
            module_name += "." + item
        class_name = comm_type.split(".")[-1]
        module = __import__("iotgate.communicators.transport" + module_name, fromlist=[class_name])

        comm_class = getattr(module, class_name)
        communicator = comm_class(comm_id, configuration)

        return communicator

    def __init__(self, id: str, type_: str, configuration: dict):
        Subject.__init__(self)
        self.id = id
        self.type = type_
        self.configuration = configuration
        self.state = "undetermined"
        self.configure()

    @abstractmethod
    def configure(self):
        pass

    def get_id(self):
        return self.id

    def _set_id(self, id_):
        self.id = id_

    def get_type(self):
        return self.type

    def get_configuration(self) -> dict:
        return self.configuration

    def set_configuration_param(self, param: str, value: str):
        self.configuration[param] = value
        self.configure()

    def get_dependent_adaptors(self) -> list:
        return self.adaptors

    def get_dependent_adaptors_id(self) -> list:
        return [a.get_id() for a in self.get_adaptors()]

    def get_state(self) -> str:
        return self.state

    def serialize(self):
        return {
            "commId": str(self.id),
            "commType": str(self.type),
            "configuration": dict(self.configuration)
        }


class CommunicatorParam(Base):
    __tablename__ = 'communicator_param'

    communicator_id = Column(String(128),
                             ForeignKey('communicator.id', onupdate='CASCADE', ondelete='CASCADE'),
                             primary_key=True)
    key = Column(String(128), primary_key=True)
    value = Column(String(128), nullable=False)

    def get_communicator_id(self):
        return self.communicator_id

    def _set_communicator_id(self, communicator_id):
        self.communicator_id = communicator_id

    def get_key(self):
        return self.key

    def _set_key(self, key):
        self.key = key

    def get_value(self):
        return self.value

    def set_value(self, value):
        self.value = value


class Server(Communicator):
    """
        Events
            server_request: called after receive any request.
                param - address => Device address information.
                param - client: Client => Client instance to communicate with device
            server_start: called after receive the node answer.
                param - configuration: str => Server's configuration.
            server_stop: called just before stop the server.
                param - message: str => Everything stop well?.
            error: called for any error occurred.
                param - message: str => Error message.
    """
    event_types = ["server_request", "server_start", "server_stop", "error"]

    def __init__(self, id: str, type_: str, configuration: dict):
        Communicator.__init__(self, id, type_, configuration)
        self.request_callbacks = []

    @abstractmethod
    def start(self):
        pass

    @abstractmethod
    def stop(self):
        pass

    def add_request_callback(self, callback):
        self.add_callback("server_request", callback)


class Client(Communicator):
    """
        Events
            client_connect: called after connect.
                param - configuration: dict => Device address information.
            client_send: called after send something.
                param - configuration: str => Client's configuration.
            client_rcv: called after receive something.
                param - configuration: str => Client's configuration.
            client_close: called just before close connection.
                param - message: str => Everything stop well?.
            info: called for any useful info for debug or log.
                param - message: str => Info message.
            error: called for any error occurred.
                param - message: str => Error message.
    """
    event_types = ["client_connect", "client_send", "client_rcv", "client_close", "info", "error"]

    @abstractmethod
    def connect(self):
        pass

    @abstractmethod
    def send(self, message):
        pass

    @abstractmethod
    def receive(self):
        pass

    @abstractmethod
    def close(self):
        pass
