import pkgutil
import inspect
import sys
import uuid

from sqlalchemy.sql.elements import between
from sqlalchemy.sql.expression import select
from sqlalchemy.sql.expression import join

from iotgate.core.model import adaptor_communicator_table
from iotgate.core.model.patterns import Subject
from iotgate.core.model.communicators import Communicator
from iotgate.core.model.communicators import CommunicatorParam
from iotgate.core.model.communicators import Client
from iotgate.core.model.communicators import Server
from iotgate.core.model.adaptors import Adaptor
from iotgate.core.model.products import Product
from iotgate.core.model.products import ProductSensor
from iotgate.core.model.products import ProductActuator
from iotgate.core.model.products import ProductActuatorAction
from iotgate.core.model.nodes import NodeGroup
from iotgate.core.model.nodes import NodeComponent
from iotgate.core.model.nodes import Node
from iotgate.core.model.nodes import Sensor
from iotgate.core.model.nodes import Actuator
from iotgate.core.model.nodes import NodeComponentState


__author__ = "diego"
__copyright__ = "Copyright 2016, Argotec SL"
__license__ = "?"
__version__ = "0.1.4-tfg"
__maintainer__ = "Diego Fidalgo"
__email__ = "dfidalgo@argotec.es"
__status__ = "Development"


class CommunicatorManager:
    """
        Events
            on_create: called after create a new node group.
                param - node_group: NodeGroup => New empty node group.
            on_delete: called after remove a node group.
                param - node_group_name: str => Removed node group unique name.
            on_change: called after add/remove nodes, or change name or description.
                param - element_changed: str => node, name or description that have changed.
                param - change_description: str => description of change.
    """
    event_types = ["on_create", "on_delete", "on_change"]

    def __init__(self, db_manager):
        self.communicators = {}
        self.db_manager = db_manager
        self.__load_communicators_from_db()

    def __load_communicators_from_db(self):
        comm_table = Communicator.__table__
        comm_param_table = CommunicatorParam.__table__

        connection = self.db_manager.engine.connect()
        try:
            for comm_raw in connection.execute(comm_table.select()):
                comm_param_stmt = select([comm_param_table]).where(comm_param_table.c.communicator_id == comm_raw['id'])
                params = {}
                for param in connection.execute(comm_param_stmt):
                    params[param['key']] = param['value']
                self.communicators[comm_raw['id']] = Communicator.create(comm_raw['id'], comm_raw['type'], params)
        finally:
            connection.close()

    @staticmethod
    def get_communicator_types() -> list:
        try:
            tr = __import__('iotgate.communicators.transport', fromlist=[''])
            module_names = [name for _, name, _ in pkgutil.iter_modules(tr.__path__)]
            comm_type_list = []
            for module_name in module_names:
                try:
                    __import__("iotgate.communicators.transport", fromlist=[module_name])
                    class_list = inspect.getmembers(sys.modules['iotgate.communicators.transport.' + module_name],
                                                    inspect.isclass)
                    for _class in class_list:
                        if _class[1] in Client.__subclasses__() or _class[1] in Server.__subclasses__():
                            comm_type_list.append(module_name + '.' + _class[0])
                except ImportError:
                    pass
            return comm_type_list
        except ImportError:
            return []

    @staticmethod
    def get_communicator_type_config_params(comm_type: str) -> list:
        module_name = ''
        for item in comm_type.split('.')[:-1]:
            module_name += '.' + item
        class_name = comm_type.split('.')[-1]
        module = __import__('iotgate.communicators.transport' + module_name, fromlist=[class_name])

        comm_class = getattr(module, class_name)

        return comm_class.CONFIGURATION_PARAMS

    @staticmethod
    def create_communicator(comm_type: str, configuration: dict):
        return Communicator.create('comm-' + str(uuid.uuid4()), comm_type, configuration)

    def get_communicators_status(self) -> dict:
        status = {}
        for comm_id in self.communicators:
            status[comm_id] = self.communicators[comm_id].serialize()
        return status

    def get_communicator_by_id(self, comm_id: str):
        return self.communicators[comm_id]

    def get_managed_communicators_ids(self):
        return self.communicators.keys()

    def start_communicator_by_id(self, comm_id: str):
        if 'start' in dir(self.communicators[comm_id]):
            self.communicators[comm_id].start()

    def stop_communicator_by_id(self, comm_id: str):
        if 'stop' in dir(self.communicators[comm_id]):
            self.communicators[comm_id].stop()

    def start_communicators(self):
        for comm_id in self.communicators:
            if 'start' in dir(self.communicators[comm_id]):
                self.communicators[comm_id].start()

    def stop_communicators(self):
        for comm_id in self.communicators:
            if 'stop' in dir(self.communicators[comm_id]):
                self.communicators[comm_id].stop()

    def restart_communicators(self):
        self.stop_communicators()
        self.start_communicators()

    def restart_communicator_by_id(self, communicator_id: str):
        self.stop_communicator_by_id(communicator_id)
        self.start_communicator_by_id(communicator_id)

    def new_communicator(self, comm_id: str, comm_type: str, configuration: dict):
        new_comm = Communicator.create(comm_id, comm_type, configuration)
        self.db_manager.session.add(new_comm)
        self.db_manager.session.commit()
        self.communicators[new_comm.get_id()] = new_comm
        return new_comm

    def remove_communicator(self, comm_id: str):
        self.db_manager.session.delete(self.communicators[comm_id])
        self.db_manager.session.commit()


class AdaptorManager(Subject):
    """
        Events
            on_create: called after create a new node group.
                param - node_group: NodeGroup => New empty node group.
            on_delete: called after remove a node group.
                param - node_group_name: str => Removed node group unique name.
            on_change: called after add/remove nodes, or change name or description.
                param - element_changed: str => node, name or description that have changed.
                param - change_description: str => description of change.
    """
    event_types = ["on_create", "on_delete", "on_change"]

    def __init__(self, db_manager, communicator_manager):
        Subject.__init__(self)
        self.db_manager = db_manager
        self.communicator_manager = communicator_manager
        self.adaptors = {}
        self.__load_adaptors_from_db()

    def __load_adaptors_from_db(self):
        assoc_table = adaptor_communicator_table

        connection = self.db_manager.engine.connect()

        try:
            for adaptor_raw in connection.execute(Adaptor.__table__.select()):
                adaptor_communicator_stmt = select([assoc_table]).where(assoc_table.c.adaptor_id == adaptor_raw['id'])
                comms = [self.communicator_manager.get_communicator_by_id(assoc_record['communicator_id'])
                         for assoc_record in connection.execute(adaptor_communicator_stmt)]
                self.adaptors[adaptor_raw['id']] = Adaptor.create(adaptor_raw['id'], adaptor_raw['type'], comms)
        finally:
            connection.close()

    def new_adaptor(self, name, adaptor_type, communicators_ids):
        communicators = [self.communicator_manager.get_communicator_by_id(_id) for _id in communicators_ids]
        adaptor = Adaptor.create(name, adaptor_type, communicators)
        self.adaptors[adaptor.get_id()] = adaptor
        self.db_manager.session.add(adaptor)
        self.db_manager.session.commit()

        return adaptor

    def remove_adaptor(self, adp_id: str):
        adp = self.adaptors[adp_id]
        self.db_manager.session.delete(adp)

    @staticmethod
    def get_adaptors_types():
        try:
            adaptor_pkg = __import__('iotgate.adaptors', fromlist=[''])
            pkg_names = [name for _, name, _ in pkgutil.walk_packages(adaptor_pkg.__path__)]
            adp_type_list = []
            for pkg_name in pkg_names:
                try:
                    pkg = __import__("iotgate.adaptors." + pkg_name, fromlist=[''])
                    modules_names = [name for _, name, _ in pkgutil.iter_modules(pkg.__path__)]
                    for module_name in modules_names:
                        try:
                            __import__("iotgate.adaptors." + pkg_name, fromlist=[module_name])
                            class_list = inspect.getmembers(sys.modules['iotgate.adaptors.' + pkg_name + '.' + module_name],
                                                            inspect.isclass)
                            for _class in class_list:
                                if _class[1] in Adaptor.__subclasses__():
                                    adp_type_list.append(pkg_name + '.' + module_name + '.' + _class[0])
                        except ImportError:
                            pass
                except ImportError:
                    pass
            return adp_type_list
        except ImportError:
            return []

    def get_manage_adaptors_names(self) -> dict:
        return list(self.adaptors.keys())

    def get_adaptors(self):
        return [self.adaptors[a] for a in self.adaptors]

    def get_adaptor_by_name(self, adaptor_name):
        return self.adaptors[adaptor_name]

    def add_node(self, node):
        self.adaptors[node.adaptor_id].add_managed_node(node)

    def delete_node(self, node):
        self.adaptors[node.get_adaptor_name()].remove_node(node)


class ProductCatalog:
    """
        Events
            on_create: called after create a new node group.
                param - node_group: NodeGroup => New empty node group.
            on_delete: called after remove a node group.
                param - node_group_name: str => Removed node group unique name.
            on_change: called after add/remove nodes, or change name or description.
                param - element_changed: str => node, name or description that have changed.
                param - change_description: str => description of change.
    """
    event_types = ["on_create", "on_delete", "on_change"]

    def __init__(self, db_manager):
        self.db_manager = db_manager
        self.products = {}
        self.__load_products_from_db()

    def __load_products_from_db(self):
        for product in self.db_manager.session.query(Product):
            self.products[product.get_id()] = product

    @staticmethod
    def __from_descriptor(product_descriptor: dict) -> Product:
        components = []
        for comp in product_descriptor["components"]:
            if comp["type"] == "sensor":
                component = ProductSensor(board_id=comp["board_id"], unit=comp["unit"], value_type=comp["value_type"])
            else:
                component = ProductActuator(board_id=comp["board_id"], unit=comp["unit"], value_type=comp["value_type"],
                                            actions=[
                                                ProductActuatorAction(name=action["name"], command=action["command"])
                                                for action in comp["actions"]
                                                ])

            components.append(component)

        product = Product(id=product_descriptor["id"],
                          manufacturer=product_descriptor["manufacturer"],
                          model=product_descriptor["model"],
                          components=components)

        return product

    def register_product(self, product_descriptor):
        product = self.__from_descriptor(product_descriptor)
        self.db_manager.session.add(product)
        self.db_manager.session.commit()
        self.products[product.get_id()] = product
        return product

    def unregister_product(self, product_id):
        pr = self.db_manager.session.query(Product).filter_by(id=product_id)[0]
        self.db_manager.session.delete(pr)
        self.db_manager.session.commit()
        del self.products[product_id]

    def get_product(self, product_id: str):
        return self.products[product_id]

    def get_products_id(self) -> list:
        return [key for key in self.products.keys()]


class NodeManager(Subject):
    """
    Events
        create_group: called after create a new node group.
            param - group: NodeGroup => New empty node group.
        delete_group: called after remove a node group.
            param - group: NodeGroup => Removed node group.
        change_group: called after change the name or description of the group.
            param - element_changed: str => node, name or description that have changed.
            param - change_description: str => description of change.
        swap_node: called after move a node between two different groups.
            param - node: Node => The moved node.
            param - from_group: NodeGroup => The node group which the node was removed.
            param - to_group: NodeGroup => The node group which the node was inserted.
    """
    event_types = ['create_group', 'delete_group', 'change_group', 'swap_node']

    def __init__(self, db_manager, adaptor_manager):
        self.db_manager = db_manager
        self.adaptor_manager = adaptor_manager
        self.node_groups = {}
        self.__load_node_groups_from_db()
        Subject.__init__(self)

    def __load_node_groups_from_db(self):
        ng_table = NodeGroup.__table__
        node_table = Node.__table__
        node_comp_table = NodeComponent.__table__

        connection = self.db_manager.engine.connect()
        try:
            for ng_raw in connection.execute(ng_table.select()):
                self.node_groups[ng_raw['id']] = NodeGroup(**ng_raw)
            # TODO se cargan los nodos pero únicamente se asocia a los adaptors, no a los grupos
            for node_raw in connection.execute(node_table.select()):
                stmt = node_comp_table.select().where(node_comp_table.c.node_id == node_raw['id'])
                comps = [NodeComponent(**node_comp_raw) for node_comp_raw in connection.execute(stmt)]
                self.adaptor_manager.add_node(Node(components=comps, **node_raw))
        finally:
            connection.close()

        for ng in self.db_manager.session.query(NodeGroup):
            self.node_groups[ng.get_id()] = ng

    def retrieve_node_groups(self) -> list:
        return self.db_manager.session.query(NodeGroup)

    def create_node_group(self, node_group_id: str, description: str) -> NodeGroup:
        if node_group_id not in self.node_groups:
            node_group = NodeGroup(id=node_group_id, description=description)
            self.db_manager.session.add(node_group)
            self.db_manager.session.commit()
            self.node_groups[node_group_id] = node_group
            self.trigger_event('create_group', group=node_group)
            return node_group
        else:
            # TODO: Raise exception, node group already exist
            raise Exception('TODO: Raise exception, node group already exist')

    def delete_node_group(self, node_group_id: str):
        node_group = self.db_manager.session.query(NodeGroup).filter_by(id=node_group_id).one()
        self.db_manager.session.delete(node_group)
        self.db_manager.session.commit()
        del self.node_groups[node_group_id]
        self.trigger_event('delete_group', group=node_group)

    def create_node(self, alias, network_id, node_group, product, adaptor, configuration={}) -> Node:
        """
        Creates a valid node and associates it to the node group.

        :param alias: the human readable id of the node.
        :type alias: str
        :param network_id: the internal id that identifies the node inside the network.
        :type network_id: str
        :param node_group: the group where the node belongs.
        :type node_group: NodeGroup
        :param product: the specification of the node.
        :type product: Product
        :param adaptor: the adaptor associated to the node.
        :type adaptor: Adaptor
        :param configuration: the internal configuration of the node.
        :type configuration: dict
        :return: the created node.
        :rtype: Node
        """
        node = Node(alias=alias, network_id=network_id, node_group=node_group, product=product, adaptor=adaptor,
                    configuration=configuration)

        for component in node.get_product().get_components():
            params = {
                'product_component': component,
                'type': component.get_type(),
                'board_id': component.get_board_id(),
                'alias': component.get_board_id()
            }

            if params['type'] == 'sensor':
                node.get_components().append(Sensor(**params))
            elif params['type'] == 'actuator':
                node.get_components().append(Actuator(**params))
            else:
                raise Exception('Invalid NodeComponent type: ' + params['type'])

        adaptor.add_managed_node(node)

        node_group.add_node(node)
        self.db_manager.session.commit()

        return node

    def delete_node(self, node_group_id: str, node_alias: str):
        """
        Removes the node from his associated node group.

        :param node_group_id: the node group which contains the node to be removed.
        :type node_group_id: str
        :param node_alias: alias from the node to be removed.
        :type node_alias: str
        """
        node_group = self.db_manager.session.query(NodeGroup).filter_by(id=node_group_id).one()
        node = node_group.get_node_by_alias(node_alias)
        node_group.remove_node(node)
        self.db_manager.session.commit()

    def swap_node(self, node: str, from_group: str, to_group: str):
        """
        Swaps the node between two different node groups.

        :param node: the alias from node which is going to be changed between groups.
        :type node: str
        :param from_group: the source node group name.
        :type from_group: str
        :param to_group: the destination node group name.
        :type to_group: str

        On success fires the event 'swap_node'.
        """
        from_group = self.db_manager.session.query(NodeGroup).filter_by(id=from_group).one()
        to_group = self.db_manager.session.query(NodeGroup).filter_by(id=to_group).one()
        from_group.nodes.delete(node)
        to_group.nodes.append(node)
        self.db_manager.session.commit()
        self.trigger_event('swap_node', node=node, from_group=from_group, to_group=to_group)

    def set_node_config(self, node_group_name, node_alias, configuration):
        self.node_groups[node_group_name].get_node_by_alias(node_alias).set_configuration(configuration)

    def get_node_groups_names(self):
        return list(self.node_groups.keys())

    def get_node_group(self, name):
        return self.node_groups[name]

    def get_nodes_groups(self):
        return [self.node_groups[ng] for ng in self.node_groups]

    def get_node_by_alias(self, node_group_id: int, node_alias: str):
        return next((node for node in self.node_groups[node_group_id].nodes
                     if node.get_alias() == node_alias), None)

    def get_node_by_id(self, node_group_id: int, node_id: int):
        return next((node for node in self.node_groups[node_group_id].nodes
                     if node.get_id() == node_id), None)

    def listen_all_groups(self, event_type: str, callback):
        """
        Binds the callback to the selected event type of all node groups.

        :param event_type: the type of the event that would like to listen (see NodeGroup entity).
        :type event_type: str
        :param callback: the callback executed when the event is fired for any node group.
        """
        for node_group_id in self.node_groups:
            self.node_groups[node_group_id].add_callback(event_type, callback)

    def listen_all_nodes_of_a_group(self, node_group_id: str, event_type: str, callback):
        """
        Binds the callback to the selected event type of all nodes of a group.

        :param node_group_id: the identifier of the node group.
        :type node_group_id: str
        :param event_type: the type of the event that would like to listen (see Node entity).
        :type event_type: str
        :param callback: the callback executed when the event is fired for any node of the selected group.
        """
        self.node_groups[node_group_id].add_callback_to_all_nodes(event_type, callback)

    def listen_all_nodes(self, event_type: str, callback):
        """
        Binds the callback to the selected event type of all nodes.

        :param event_type: the type of the event that would like to listen (see Node entity).
        :type event_type: str
        :param callback: the callback executed when the event is fired for any node.
        """
        for ng_name in self.node_groups:
            self.node_groups[ng_name].add_callback_to_all_nodes(event_type, callback)

    def get_data(self, node_group_id: str, node_id: int, board_id: str, from_date, to_date) -> list:
        '''
        Fetch the data of the selected component
        :param node_group_id:
        :param node_id:
        :param board_id:
        :param from_date:
        :param to_date:
        :return:
        '''
        node_stream_table = NodeComponent.__table__
        node_state_table = NodeComponentState.__table__

        connection = self.db_manager.engine.connect()

        j = join(node_stream_table, node_state_table, node_stream_table.c.id == node_state_table.c.stream_id)
        stmt = select([node_state_table]).select_from(j)\
            .where(node_stream_table.c.board_id == board_id)\
            .where(node_stream_table.c.node_id == node_id)\
            .where(between(node_state_table.c.observed_at, from_date, to_date))

        return [NodeComponentState(**state_row) for state_row in connection.execute(stmt)]

    def get_last_data(self, node_group_id: str, node_id: int, board_id: str) -> NodeComponentState:
        '''
        Fetch the last data of the selected component
        :param node_group_id:
        :param node_id:
        :param board_id:
        :return:
        '''
        node_stream_table = NodeComponent.__table__
        node_state_table = NodeComponentState.__table__

        j = join(node_stream_table, node_state_table, node_stream_table.c.id == node_state_table.c.stream_id)
        stmt = select([node_state_table]).select_from(j) \
            .where(node_stream_table.c.board_id == board_id) \
            .where(node_stream_table.c.node_id == node_id) \
            .order_by(node_state_table.c.observed_at).limit(1)

        connection = self.db_manager.engine.connect()
        try:
            state_row = connection.execute(stmt).fetchone()
        finally:
            connection.close()

        return NodeComponentState(**state_row) if state_row else None
