from iotgate.core.model.products import Product
from sqlalchemy.ext.associationproxy import association_proxy
from sqlalchemy.orm import relationship
from sqlalchemy.orm import reconstructor
from sqlalchemy.orm.collections import attribute_mapped_collection
from sqlalchemy.sql.schema import Table
from sqlalchemy.sql.schema import Column
from sqlalchemy.sql.schema import UniqueConstraint
from sqlalchemy.sql.schema import ForeignKey
from sqlalchemy.sql.sqltypes import Integer
from sqlalchemy.sql.sqltypes import Boolean
from sqlalchemy.sql.sqltypes import String
from sqlalchemy.sql.sqltypes import DateTime

from iotgate.core.model import Base
from iotgate.core.model.patterns import Subject
from datetime import datetime

__author__ = 'diego'
__copyright__ = 'Copyright 2016, Argotec SL'
__license__ = '?'
__version__ = '0.1.2'
__maintainer__ = 'Diego Fidalgo'
__email__ = 'dfidalgo@argotec.es'
__status__ = 'Development'

node_group_node_user_table = Table('node_group_node_user', Base.metadata,
                                   Column('node_group_id', String(128), ForeignKey('node_group.id')),
                                   Column('node_user_id', Integer, ForeignKey('node_user.id'))
                                   )


class NodeUser(Base):
    __tablename__ = 'node_user'

    id = Column('id', Integer, primary_key=True, autoincrement=True)
    password = Column('password', String(128))
    last_login = Column('last_login', DateTime(True))
    is_superuser = Column('is_superuser', Boolean)
    username = Column('username', String(30), unique=True)
    first_name = Column('first_name', String(30))
    last_name = Column('last_name', String(30))
    email = Column('email', String(30))
    is_staff = Column('is_staff', Boolean)
    is_active = Column('is_active', Boolean)
    date_joined = Column('date_joined', DateTime(True))

    groups = relationship('NodeGroup', secondary=node_group_node_user_table)

    @property
    def is_authenticated(self):
        return True

    @property
    def is_anonymous(self):
        return False

    def get_id(self):
        return self.username

    def get_simple_info(self) -> dict:
        return {
            'username': self.username,
            'firstName': self.first_name,
            'lastName': self.last_name,
            'email': self.email
        }


class NodeGroup(Base, Subject):
    '''
    Events
        change: called after change an owner field of the NodeGroup.
            param - type: str => The name of the field changed.
            param - old_value => The previous value of the field.
            param - new_value => The new value of the field.
        node_added: called after add a node into a group.
            param - node: Node => The new node.
            param - group: NodeGroup => The node group which the node was added.
        node_removed: called after remove a node from a group.
            param - node: Node => The removed node.
            param - group: NodeGroup => The node group which the node was removed.
    '''
    event_types = ['change', 'node_added', 'node_removed']

    __tablename__ = 'node_group'

    id = Column(String(128), primary_key=True)
    description = Column(String(256))
    register_date = Column(DateTime(True), default=datetime.now, nullable=False)
    nodes = relationship('Node', lazy='subquery')

    def __init__(self, **kwargs):
        Base.__init__(self, **kwargs)
        self._load_to_init()

    @reconstructor
    def _load_to_init(self):
        Subject.__init__(self)

    def get_id(self):
        return self.id

    def _set_id(self, id_):
        self.id = id_

    def get_description(self):
        return self.description

    def set_description(self, description):
        self.trigger_event('change', type='description', old_value=self.description, new_value=description)
        self.description = description

    def get_register_date(self):
        return self.register_date

    def set_register_date(self, register_date):
        self.trigger_event('change', type='register_date', old_value=self.register_date, new_value=register_date)
        self.register_date = register_date

    def add_node(self, node):
        self.nodes.append(node)
        self.trigger_event('node_added', node=node, group=self)

    def remove_node(self, node):
        self.nodes.delete(node)
        self.trigger_event('node_removed', node=node, group=self)

    def get_nodes(self):
        return self.nodes

    def get_node_by_id(self, id_):
        for node in self.nodes:
            if node.get_id() == id_:
                return node
        # TODO: Raise node not exist

    def get_nodes_id(self):
        return [node.get_id() for node in self.nodes]

    def get_node_by_alias(self, alias):
        for node in self.nodes:
            if node.get_alias() == alias:
                return node
        # TODO: Raise node not exist

    def get_nodes_alias(self):
        return [node.get_alias() for node in self.nodes]

    def add_callback_to_all_nodes(self, event_type: str, callback):
        for node in self.nodes:
            node.add_callback(event_type, callback)

    def serialize(self, nodes=False, **kwargs):
        return {
            'id': self.id,
            'description': self.description,
            'registeredAt': str(self.register_date),
            'nodes': [node.serialize(**kwargs) for node in self.nodes] if nodes else None
        }


class Node(Base, Subject):
    '''
    Events
        new_state: called after receive a new state.
            param - component: NodeComponent => The component.
            param - data: NodeComponentState => The state received.
        change: called after change an owner field of the node.
            param - type: str => The name of the field changed.
            param - old_value => The previous value of the field.
            param - new_value => The new value of the field.
        component_added: called after plug a component to the node.
            param - component: NodeComponent => The component connected to the node.
            param - node: Node => The current node.
        component_removed: called after unplug a component from the node.
            param - component: NodeComponent => The component unconnected from the node.
            param - node: Node => The current node.
        component_swapped: called after swap a component in the same node.
            param - component: NodeComponent => The component swapped.
            param - old_board_id: str => The previous board identifier of the component in the node.
            param - new_board_id: str => The new board identifier of the component in the node.
            param - node: Node => The current node.
    '''
    event_types = ['new_state', 'change', 'component_added', 'component_removed', 'component_swapped']

    __tablename__ = 'node'

    id = Column(Integer, autoincrement=True, primary_key=True)
    product_id = Column(String(128),
                        ForeignKey('product.id', onupdate='CASCADE', ondelete='SET NULL'))
    product = relationship('Product')
    node_group_id = Column(String(128),
                           ForeignKey('node_group.id', onupdate='CASCADE', ondelete='CASCADE'))
    node_group = relationship('NodeGroup')
    adaptor_id = Column(String(128), ForeignKey('adaptor.id',
                                                onupdate='CASCADE', ondelete='SET NULL'))
    adaptor = relationship('Adaptor')
    params = relationship('NodeConfigurationParam', collection_class=attribute_mapped_collection('key'))
    configuration = association_proxy('params', 'value', creator=lambda k, v: NodeConfigurationParam(key=k, value=v))
    components = relationship('NodeComponent', lazy='subquery')
    network_id = Column(String(128), nullable=False)
    alias = Column(String(128), nullable=False)

    # Node alias must be unique under a concrete node group
    UniqueConstraint(node_group_id, alias, name='unique_node_in_ngroup')

    def __init__(self, **kwargs):
        Base.__init__(self, **kwargs)
        Subject.__init__(self)
        self._init_on_load()

    @reconstructor
    def _init_on_load(self):
        Subject.__init__(self)

    def get_id(self) -> str:
        return self.id

    def get_product(self) -> Product:
        return self.product

    def get_node_group(self) -> NodeGroup:
        return self.node_group

    def set_node_group(self, node_group: NodeGroup):
        self.trigger_event('change', type='node_group', old_value=self.node_group, new_value=node_group)
        self.node_group = node_group

    def get_adaptor(self):
        return self.adaptor

    def get_configuration(self) -> dict:
        return self.configuration

    def set_configuration(self, configuration: dict):
        self.trigger_event('change', type='configuration', old_value=self.configuration, new_value=configuration)
        self.configuration = configuration

    def get_network_id(self) -> str:
        return self.network_id

    def set_network_id(self, network_id: str):
        self.trigger_event('change', type='network_id', old_value=self.network_id, new_value=network_id)
        self.network_id = network_id

    def get_alias(self) -> str:
        return self.alias

    def set_alias(self, alias: str):
        self.trigger_event('change', type='alias', old_value=self.alias, new_value=alias)
        self.alias = alias

    def add_component(self, component):
        '''
        Adds or 'plug' the component in the current node.

        :param component: the added component in the node.
        :type component: NodeComponent

        Fires the event 'component_added'.
        '''
        self.components.append(component)
        self.trigger_event('component_added', component=component, node=self)

    def remove_component(self, component):
        '''
        Removes or 'unplug' the component from the current node.

        :param component: the removed component from the node.
        :type component: NodeComponent

        Fires the event 'component_removed'.
        '''
        self.components.remove(component)
        self.trigger_event('component_removed', component=component, node=self)

    def swap_component(self, component, new_board_id: str):
        '''
        Swaps the component between two plugs of the same node.

        :param component: the swapped component.
        :type component: NodeComponent
        :param new_board_id: the identifier of the plug where the component will be connected.
        :type new_board_id: str

        Fires the event 'swap_component'.
        '''
        old_board_id = component.get_board_id()
        component.set_board_id(new_board_id)
        self.trigger_event('component_swapped', component=component, old_board_id=old_board_id,
                           new_board_id=new_board_id, node=self)

    def get_components(self) -> list:
        return self.components

    def get_sensors(self) -> list:
        return [component for component in self.components if isinstance(component, Sensor)]

    def get_actuators(self) -> list:
        return [component for component in self.components if isinstance(component, Actuator)]

    def get_component_by_alias(self, alias: str):
        return next((s for s in self.components if s.get_alias() == alias))

    def get_component_by_board_id(self, board_id: str):
        return next(s for s in self.components if s.get_board_id() == board_id)

    def serialize(self, components=False):
        return {
            'id': self.id,
            'alias': self.alias,
            'nodeGroupId': self.node_group_id,
            'networkId': self.network_id,
            'productId': self.product_id,
            'adaptorId': self.adaptor_id,
            'configuration': self.configuration,
            'components': [component.serialize() for component in self.components] if components else []
        }


class NodeConfigurationParam(Base):
    __tablename__ = 'node_configuration_param'

    node_id = Column(Integer, ForeignKey('node.id', onupdate='CASCADE', ondelete='CASCADE'), primary_key=True)
    key = Column(String(128), primary_key=True)
    value = Column(String(512), nullable=False)

    def get_node_id(self):
        return self.node_id

    def _set_node_id(self, node_id):
        self.node_id = node_id

    def get_key(self):
        return self.key

    def _set_key(self, key):
        self.key = key

    def get_value(self):
        return self.value

    def set_value(self, value):
        self.value = value


class NodeComponent(Base):
    __tablename__ = 'node_component_stream'

    id = Column(Integer, autoincrement=True, primary_key=True)
    product_component_id = Column(String(128),
                                  ForeignKey('product_component.id',
                                             onupdate='CASCADE', ondelete='SET NULL'))
    product_component = relationship('ProductComponent')
    board_id = Column(String(128), nullable=False)
    active = Column(Boolean(), default=True, nullable=False)
    node_id = Column(Integer,
                     ForeignKey('node.id', onupdate='CASCADE', ondelete='CASCADE'),
                     nullable=False)
    node = relationship('Node')
    states = relationship('NodeComponentState', lazy='dynamic')
    alias = Column(String(128))

    type = Column(String(128), nullable=False)
    __mapper_args__ = {'polymorphic_on': type}

    def __init__(self, **kwargs):
        Base.__init__(self, **kwargs)
        self.last_state = None

    def get_id(self):
        return self.id

    def _set_id(self, id_):
        self.id = id_

    def get_type(self):
        return self.type

    def _set_type(self, type_):
        self.type = type_

    def get_product_component_id(self):
        return self.product_component_id

    def _set_product_component_id(self, product_component_id):
        self.product_component_id = product_component_id

    def get_product_component(self):
        return self.product_component

    def get_unit(self):
        return self.product_component.get_unit()

    def set_product_component(self, product_component):
        self.product_component = product_component

    def get_board_id(self):
        return self.board_id

    def set_board_id(self, board_id):
        self.board_id = board_id

    def is_active(self):
        return self.active

    def set_active(self, active):
        self.active = active

    def get_node_id(self):
        return self.node_id

    def _set_node_id(self, node_id):
        self.node_id = node_id

    def get_node(self):
        return self.node

    def set_node(self, node):
        self.node = node

    def get_states(self):
        return self.states

    def get_alias(self):
        return self.alias

    def set_alias(self, alias):
        self.alias = alias

    def get_last_state(self):
        return self.last_state

    def set_last_state(self, new_state):
        self.last_state = new_state
        self.get_node().trigger_event('new_state', component=self, data=new_state)

    def serialize(self):
        return {
            'id': self.id,
            'alias': self.alias,
            'type': self.type,
            'boardId': self.board_id,
            'nodeId': self.node_id,
            'productComponentId': self.product_component_id,
            'active': self.active
        }


class Sensor(NodeComponent):
    __mapper_args__ = {'polymorphic_identity': 'sensor'}


class Actuator(NodeComponent):
    __mapper_args__ = {'polymorphic_identity': 'actuator'}

    def get_actions(self) -> list:
        return self.get_product_component().get_actions()

    def do_action(self, name, params):
        action = self.get_product_component().get_action(name)
        new_state = self.get_node().get_adaptor().send_action(self, action, params)
        self.set_last_state(new_state)

        return new_state


class NodeComponentState(Base):
    __tablename__ = 'node_component_state'

    stream_id = Column(Integer,
                       ForeignKey('node_component_stream.id', onupdate='CASCADE', ondelete='CASCADE'),
                       primary_key=True)
    stream = relationship('NodeComponent')
    observed_at = Column(DateTime(True), default=datetime.now, primary_key=True)
    received_at = Column(DateTime(True), default=datetime.now, nullable=False)
    value = Column(String(512), nullable=False)

    def get_stream_id(self):
        return self.stream_id

    def _set_stream_id(self, stream_id):
        self.stream_id = stream_id

    def get_stream(self):
        return self.stream

    def set_stream(self, stream):
        self.stream = stream

    def get_observed_at(self):
        return self.observed_at

    def _set_observed_at(self, observed_at):
        self.observed_at = observed_at

    def get_received_at(self):
        return self.received_at

    def set_received_at(self, received_at):
        self.received_at = received_at

    def get_value(self):
        return self.value

    def set_value(self, value):
        self.value = value

    def serialize(self):
        return {
            'componentId': self.stream_id,
            'observedAt': str(self.observed_at),
            'receivedAt': str(self.received_at),
            'value': self.value
        }
