from sqlalchemy.sql.schema import Column
from sqlalchemy.sql.schema import ForeignKey
from sqlalchemy.sql.sqltypes import String
from sqlalchemy.sql.sqltypes import Integer
from sqlalchemy.orm import relationship

from iotgate.core.model import Base
from iotgate.core.model import product_component_table
from iotgate.core.model.patterns import Subject

__author__ = "diego"
__copyright__ = "Copyright 2016, Argotec SL"
__license__ = "?"
__version__ = "0.1.4-tfg"
__maintainer__ = "Diego Fidalgo"
__email__ = "dfidalgo@argotec.es"
__status__ = "Development"


class Product(Base, Subject):
    __tablename__ = 'product'

    id = Column(String(128), primary_key=True)
    manufacturer = Column(String(128), nullable=False)
    model = Column(String(128), nullable=False)
    components = relationship('ProductComponent', secondary=product_component_table)

    def get_id(self):
        return self.id

    def _set_id(self, id_):
        self.id = id_

    def get_manufacturer(self):
        return self.manufacturer

    def set_manufacturer(self, manufacturer):
        self.manufacturer = manufacturer

    def get_model(self):
        return self.model

    def set_model(self, model):
        self.model = model

    def get_components(self):
        return self.components

    def get_component_by_board(self, board_id):
        return next((c for c in self.get_components() if c.get_board_id() == board_id), None)

    def get_components_board_id(self):
        return [c.get_board_id() for c in self.get_components()]

    def serialize(self):
        return {
            'id': self.id,
            'manufacturer': self.manufacturer,
            'model': self.model,
            'components': [component.serialize() for component in self.components]
        }


class ProductComponent(Base):
    __tablename__ = 'product_component'

    id = Column(Integer, autoincrement=True, primary_key=True)
    products = relationship('Product', secondary=product_component_table)
    actions = relationship('ProductActuatorAction')
    type = Column(String(128), nullable=False)
    __mapper_args__ = {'polymorphic_on': type}
    unit = Column(String(128), nullable=True)
    value_type = Column(String(128), nullable=True)
    board_id = Column(String(128), nullable=False)

    def get_id(self):
        return self.id

    def _set_id(self, id_):
        self.id = id_

    def _get_product_id(self):
        return self.product_id

    def _set_product_id(self, product_id):
        self.product_id = product_id

    def get_products(self):
        return self.products

    def get_type(self):
        return self.type

    def set_type(self, type_):
        self.type = type_

    def get_unit(self):
        return self.unit

    def set_unit(self, unit):
        self.unit = unit

    def get_value_type(self):
        return self.value_type

    def set_value_type(self, value_type):
        self.value_type = value_type

    def get_board_id(self):
        return self.board_id

    def set_board_id(self, board_id):
        self.board_id = board_id

    def serialize(self):
        return {
            'id': self.id,
            'type': self.type,
            'unit': self.unit,
            'valueType': self.value_type,
            'defaultBoardId': self.board_id
        }


class ProductSensor(ProductComponent):
    __mapper_args__ = {'polymorphic_identity': 'sensor'}


class ProductActuator(ProductComponent):
    __mapper_args__ = {'polymorphic_identity': 'actuator'}

    def get_actions(self):
        return self.actions

    def get_action(self, name):
        return next((a for a in self.get_actions() if a.get_name() == name), None)

    def get_actions_names(self):
        return list(self.actions.keys())

    def serialize(self):
        return ProductComponent.serialize(self).update({
            'actions': [action.serialize() for action in self.actions]
        })


class ProductActuatorAction(Base):
    __tablename__ = 'product_actuator_action'

    name = Column(String(128), nullable=True, primary_key=True)
    product_component_id = Column(Integer,
                                  ForeignKey('product_component.id', onupdate='CASCADE', ondelete='CASCADE'),
                                  primary_key=True)
    product_component = relationship('ProductComponent')
    command = Column(String(128), nullable=True)
    effect = Column(String(128), nullable=True)

    def get_name(self):
        return self.name

    def _set_name(self, name):
        self.name = name

    def _get_product_component_id(self):
        return self.product_component_id

    def _set_product_component_id(self, product_component_id):
        self.product_component_id = product_component_id

    def get_product_component(self):
        return self.product_component

    def set_product_component(self, product_component):
        self.product_component = product_component

    def get_command(self):
        return self.command

    def set_command(self, command):
        self.command = command

    def get_effect(self):
        return self.effect

    def set_effect(self, effect):
        self.effect = effect

    def serialize(self):
        return {
            'name': self.name,
            'product_component_id': self.product_component_id,
            'command': self.command,
            'effect': self.effect
        }
