import collections
import json
from iotgate.deployment.serializer import SerializerProxy
from iotgate.deployment.serializer import NodeSerializer
from iotgate.core.model.patterns import Subject
from abc import abstractmethod

__author__ = "diego"
__copyright__ = "Copyright 2016, Argotec SL"
__license__ = "?"
__version__ = "0.1.4-tfg"
__maintainer__ = "Diego Fidalgo"
__email__ = "dfidalgo@argotec.es"
__status__ = "Development"


class Publisher:
    def __init__(self, engine):
        engine.get_node_manager().add_callback_to_all_nodes(self._publish_new_values)

    def _publish_new_values(self, component, data):
        # In this case, emitter is a component
        node_group = component.get_node().get_node_group().get_id()
        node_alias = component.get_node().get_alias()
        node_component = component.get_alias()
        topic = "new_state" + "." + node_group + "." + node_alias + "." + node_component
        value = NodeSerializer.component_value_to_dict(data)
        self.publish(topic, value)

    @abstractmethod
    def publish(self, topic, serial_value):
        pass


class RequestProxy(Subject):

    """
    Events
        on_message: called after receive a message.
            param - message: str => The message received.
        on_error: called for any error occurred.
            param - message: str => Error message.
    """
    event_types = ["on_message", "on_error"]

    def __init__(self, engine):
        Subject.__init__(self)
        self.workers = engine.get_workers()

    def do_request(self, to_manager: str, operation: str, parameters: dict) -> dict:
        try:
            reply = self.workers[to_manager].do_request(operation, parameters)
            self.trigger_event("on_message", message="Received request to "
                                                     + to_manager + ", operation <" + operation + ">,"
                               "parameters: " + json.dumps(parameters, indent=3))
            return reply
        except Exception as err:
            message = "Error: something terrible happens -> " + str(err)
            self.trigger_event("on_error", message="error:::" + message)
            return {
                    "from": self.__class__.__name__,
                    "return": "Error: something terrible happens: " + str(err),
                    "operation": "None"
            }


class RequestReplyHelper:
    def __init__(self, an_object):
        self.an_object = an_object
        self.available_operations = [method for method in dir(self.an_object)
                                     if isinstance(getattr(self.an_object, method),
                                        collections.Callable) and method[0] != "_"]
        self.operations_desc = self._create_operations_descriptor()

    def _create_operations_descriptor(self):
        operations = {}
        for operation in self.available_operations:
            if operation == 'db_session':
                continue
            param_list = list(getattr(self.an_object, operation).__code__.co_varnames)
            param_list = param_list[:getattr(self.an_object, operation).__code__.co_argcount]
            try:
                param_list.remove('self')
            except ValueError:
                pass
            operations[operation] = param_list

        return operations

    def _create_serial_response(self, operation, parameters, response):
        serial_response = {
            "from": self.__class__.__name__,
            "parameters": parameters,
            "operation": operation,
            "return": SerializerProxy.serialize(response)
        }

        return serial_response

    def do_request(self, operation, parameters):
        if operation in self.available_operations:
            try:
                response = getattr(self.an_object, operation)(**parameters)
                reply = self._create_serial_response(operation, parameters, response)
            except KeyError as err:
                response = "Exception (Does not exists): " + str(err)
                reply = self._create_serial_response(operation,  parameters, response)
            except TypeError as err:
                reply = self._create_serial_response(operation,
                                                     parameters,
                                                     'Not recognized operation: invalid parameters <' + str(err) + '>')
        elif operation == "get_operations":
            reply = self._create_serial_response(operation, {}, self.operations_desc)
        else:
            reply = self._create_serial_response(operation, {}, "Requested operation doesn't exists. Privileges?")
        return reply


class ProductCatalogHelper(RequestReplyHelper):
    def __init__(self, product_catalog):
        RequestReplyHelper.__init__(self, product_catalog)


class NodeManagerHelper(RequestReplyHelper):
    def __init__(self, node_manager):
        RequestReplyHelper.__init__(self, node_manager)


class AdaptorManagerHelper(RequestReplyHelper):
    def __init__(self, adaptor_manager):
        RequestReplyHelper.__init__(self, adaptor_manager)


class CommunicatorManagerHelper(RequestReplyHelper):
    def __init__(self, communicator_manager):
        RequestReplyHelper.__init__(self, communicator_manager)
