from iotgate.core.model.adaptors import Adaptor
from iotgate.core.model.nodes import NodeGroup
from iotgate.core.model.communicators import Communicator
from iotgate.core.model.products import Product
from iotgate.core.model.products import ProductSensor
from iotgate.core.model.products import ProductActuator
from iotgate.core.model.products import ProductActuatorAction

__author__ = "diego"
__copyright__ = "Copyright 2016, Argotec SL"
__license__ = "?"
__version__ = "0.1.4-tfg"
__maintainer__ = "Diego Fidalgo"
__email__ = "dfidalgo@argotec.es"
__status__ = "Development"


class SerializerProxy(object):
    @staticmethod
    def serialize(an_object):
        if isinstance(an_object, str):
            return an_object
        if isinstance(an_object, int):
            return an_object
        if isinstance(an_object, float):
            return an_object
        if isinstance(an_object, bytes):
            return int(an_object)
        if not an_object:
            return None
        if isinstance(an_object, list):
            return [SerializerProxy.serialize(item) for item in an_object]
        if isinstance(an_object, dict):
            serial_dict = {}
            for key in an_object:
                serial_dict[key] = SerializerProxy.serialize(an_object[key])
            return serial_dict
        if isinstance(an_object, NodeGroup):
            return NodeSerializer.node_group_to_dict(an_object)
        if isinstance(an_object, Product):
            return ProductSerializer.product_to_dict(an_object)
        if isinstance(an_object, Adaptor):
            return AdaptorSerializer.adaptor_to_dict(an_object)
        if isinstance(an_object, Communicator):
            return CommunicatorSerializer.communicator_to_dict(an_object)


class ProductSerializer(object):
    @staticmethod
    def product_to_dict(product):
        product_descriptor = {
            "model": product.get_model(),
            "id": product.get_id(),
            "manufacturer": product.get_manufacturer(),
            "components": [ProductSerializer.product_component_to_dict(component)
                           for component in product.get_components()]
        }
        return product_descriptor

    @staticmethod
    def product_component_to_dict(product_component):
        component_descriptor = {
            "board_id": product_component.get_board_id(),
        }
        component_type = 'unspecified'
        if product_component.__class__.__name__ == 'ProductSensor':
            component_type = 'sensor'
            component_descriptor["unit"] = product_component.get_unit()
            component_descriptor["value_type"] = product_component.get_value_type()
        if product_component.__class__.__name__ == 'ProductActuator':
            component_type = 'actuator'
            component_descriptor["unit"] = product_component.get_unit()
            component_descriptor["value_type"] = product_component.get_value_type()
            actions = [ProductSerializer.component_action_to_dict(action)
                       for action in product_component.get_actions()]
            component_descriptor["actions"] = actions
        component_descriptor["type"] = component_type

        return component_descriptor

    @staticmethod
    def component_action_to_dict(action):
        action_descriptor = {
            "name": action.get_name(),
            "command": action.get_command(),
            "effect": action.get_effect()
        }

        return action_descriptor

    @staticmethod
    def from_descriptor(product_descriptor: dict) -> Product:
        components = []
        for comp in product_descriptor["components"]:
            if comp["type"] == "sensor":
                component = ProductSensor(board_id=comp["board_id"], unit=comp["unit"], value_type=comp["value_type"])
            else:
                component = ProductActuator(board_id=comp["board_id"], unit=comp["unit"], value_type=comp["value_type"],
                                            actions=[
                                                ProductActuatorAction(name=action["name"], command=action["command"])
                                                for action in comp["actions"]
                                                ])

            components.append(component)

        product = Product(id=product_descriptor["id"],
                          manufacturer=product_descriptor["manufacturer"],
                          model=product_descriptor["model"],
                          components=components)

        return product


class NodeSerializer(object):
    @staticmethod
    def node_group_to_dict(node_group):
        node_group_descriptor = {
            "name": node_group.get_id(),
            "description": node_group.get_description(),
            "registry_date": str(node_group.get_register_date()),
            "nodes": [NodeSerializer.node_to_dict(node) for node in node_group.get_nodes()]
        }

        return node_group_descriptor

    @staticmethod
    def node_to_dict(node):
        node_descriptor = {
            "alias": node.get_alias(),
            "product": ProductSerializer.product_to_dict(node.get_product()),
            # "configuration": node.get_configuration(),
            "network_id": node.get_network_id(),
            "adaptor_id": node.get_adaptor().get_id(),
            "node_group": node.get_node_group().get_id(),
            "components": {
                "sensors": [NodeSerializer.sensor_to_dict(sensor) for sensor in node.get_sensors()],
                "actuators": [NodeSerializer.actuator_to_dict(actuator) for actuator in node.get_actuators()],
            }
        }

        return node_descriptor

    @staticmethod
    def component_to_dict(component):
        component_descriptor = {
            "node_alias": component.get_node().get_alias(),
            "alias": component.get_alias(),
            "board_id": component.get_board_id(),
            "unit": component.get_unit()
        }
        return component_descriptor

    @staticmethod
    def unspecified_to_dict(unspecified_component):
        return ProductSerializer.product_component_to_dict(unspecified_component)

    @staticmethod
    def component_value_to_dict(component_value):
        if component_value is None:
            return {
                "received_at": "Not received yet",
                "observed_at": "Not received yet",
                "value": "Not received yet"
            }
        cv_descriptor = {
            "received_at": str(component_value.get_received_at()),
            "observed_at": str(component_value.get_observed_at()),
            "value": component_value.get_value()
        }

        return cv_descriptor

    @staticmethod
    def sensor_to_dict(sensor):
        sensor_descriptor = NodeSerializer.component_to_dict(sensor)
        sensor_descriptor["state"] = NodeSerializer.component_value_to_dict(sensor.get_last_state())

        return sensor_descriptor

    @staticmethod
    def actuator_to_dict(actuator):
        actuator_descriptor = NodeSerializer.component_to_dict(actuator)
        actuator_descriptor["state"] = NodeSerializer.component_value_to_dict(actuator.get_last_state())
        actuator_descriptor["actions"] = actuator.get_actions()

        return actuator_descriptor


class AdaptorSerializer(object):
    @staticmethod
    def adaptor_to_dict(adaptor):
        adaptor_descriptor = {
            "name": adaptor.get_id(),
            "type": adaptor.get_type(),
            "managed_nodes": adaptor.get_managed_nodes_networks_ids()
        }
        return adaptor_descriptor


class CommunicatorSerializer(object):
    @staticmethod
    def communicator_to_dict(communicator):
        communicator_descriptor = {
            "name": communicator.get_id(),
            "type": communicator.get_type(),
            "is_working": communicator.get_state(),
            "configuration": communicator.get_configuration()
        }
        return communicator_descriptor
